import os
import re

from common import TranslationJob, nice_mkdir, nice_path_join

"""
    Types of input lines
"""
TYPE_WHITESPACE = 1 # whitespace only
TYPE_HEADER     = 2 # header (beginning with \b or \t)
TYPE_BULLET     = 3 # bullet point
TYPE_IMAGE      = 4 # image (beginning with \image)
TYPE_CODE       = 5 # code (beginning with \s;)
TYPE_PLAIN      = 6 # plain text

class HelpTranslationJob(TranslationJob):
    def __init__(self, **kwargs):
        TranslationJob.__init__(self, **kwargs)
        self.template_file = kwargs['template_file']
        self.language_file = kwargs['language_file']
        self.line_buffer = None

    def process_file(self):
        while True:
            (paragraph, paragraph_type) = self.read_input_paragraph()
            if not paragraph:
                break

            if paragraph_type == TYPE_WHITESPACE:
                self.process_whitespace(paragraph[0])
            elif paragraph_type == TYPE_HEADER:
                self.process_header(paragraph[0])
            elif paragraph_type == TYPE_BULLET:
                self.process_bullet(paragraph[0])
            elif paragraph_type == TYPE_IMAGE:
                self.process_image(paragraph[0])
            elif paragraph_type == TYPE_CODE:
                self.process_code(paragraph)
            elif paragraph_type == TYPE_PLAIN:
                self.process_plain(paragraph)

    """
        Read one or more lines of input with same line type and return the list as paragraph
        Exception is types which are processed as single lines, giving only paragraph with one line
    """
    def read_input_paragraph(self):
        paragraph = None
        paragraph_type = None
        while True:
            line = None
            line_type = None
            if self.line_buffer:
                (line, line_type) = self.line_buffer
                self.line_buffer = None
            else:
                line = self.read_input_line()
                if line:
                    line_type = self.check_line_type(line.text)

            if not line:
                break

            if not paragraph_type:
                paragraph_type = line_type

            if paragraph_type == line_type:
                if not paragraph:
                    paragraph = []
                paragraph.append(line)
            else:
                self.line_buffer = (line, line_type)
                break

            if line_type in [TYPE_WHITESPACE, TYPE_HEADER, TYPE_BULLET, TYPE_IMAGE]:
                break

        return (paragraph, paragraph_type)

    def check_line_type(self, line):
        if re.match(r'^\s*$', line) or re.match(r'^\\[nctr];$', line):
            return TYPE_WHITESPACE
        elif re.match(r'^\\[bt];', line):
            return TYPE_HEADER
        elif re.match(r'^\s*([0-9]\)|[o-])', line):
            return TYPE_BULLET
        elif re.match(r'^\\image.*;$', line):
            return TYPE_IMAGE
        elif re.match(r'^\\s;', line):
            return TYPE_CODE
        else:
            return TYPE_PLAIN

    def process_whitespace(self, line):
        self.write_output_line(line.text)

    def process_header(self, line):
        match = re.match(r'^(\\[bt];)(.*)', line.text)
        header_type = match.group(1)
        header_text = match.group(2)
        translated_header_text = self.translate_text(header_text, line.occurrence, header_type + ' header')
        self.write_output_line(header_type + translated_header_text)

    def process_bullet(self, line):
        match = re.match(r'^(\s*)([0-9]\)|[o-])(\s*)(.*)', line.text)
        spacing_before_bullet = match.group(1)
        bullet_point          = match.group(2)
        spacing_after_bullet  = match.group(3)
        text                  = match.group(4)
        translated_text = self.translate_text(
            text, line.occurrence, "Bullet: '{0}'".format(bullet_point))
        self.write_output_line(spacing_before_bullet + bullet_point + spacing_after_bullet + translated_text)

    def process_image(self, line):
        match = re.match(r'^(\\image )(.*)( \d* \d*;)$', line.text)
        image_command = match.group(1)
        image_source = match.group(2)
        image_coords = match.group(3)
        translated_image_source = self.translate_text(image_source, line.occurrence, 'Image filename')
        self.write_output_line(image_command + translated_image_source + image_coords)

    def process_code(self, paragraph):
        text_lines = []
        for line in paragraph:
            match = re.match(r'^\\s;(.*)', line.text)
            code_line = match.group(1)
            text_lines.append(code_line)

        joined_text_lines = '\n'.join(text_lines)
        translated_text_lines = self.translate_text(joined_text_lines, paragraph[0].occurrence, 'Source code')
        for line in translated_text_lines.split('\n'):
            self.write_output_line(r'\s;' + line)

    def process_plain(self, paragraph):
        text_lines = []
        for line in paragraph:
            text_lines.append(line.text)

        joined_text_lines = '\n'.join(text_lines)
        translated_text_lines = self.translate_text(joined_text_lines, paragraph[0].occurrence, 'Plain text')
        for line in translated_text_lines.split('\n'):
            self.write_output_line(line)

    def translate_text(self, text, occurrence, type_comment):
        converted_text = convert_escape_syntax_to_tag_syntax(text)
        self.template_file.insert_entry(converted_text, occurrence, type_comment)

        if not self.language_file:
            return text

        translated_text = self.language_file.translate(converted_text)
        return convert_tag_syntax_to_escape_syntax(translated_text)

def convert_escape_syntax_to_tag_syntax(text):
    # Replace \button $id; as pseudo xHTML <button $id/> tags
    text = re.sub(r'\\(button|key) ([^;]*?);', r'<\1 \2/>', text)
    # Put \const;Code\norm; sequences into pseudo-HTML <format const> tags
    text = re.sub(r'\\(const|type|token|key);([^\\;]*?)\\norm;', r'<format \1>\2</format>', text)
    # Transform CBot links \l;text\u target; into pseudo-HTML <a target>text</a>
    text = re.sub(r'\\l;(.*?)\\u (.*?);', r'<a \2>\1</a>', text)
    # Cleanup pseudo-html targets separated by \\ to have a single character |
    text = re.sub(r'<a (.*?)\\(.*?)>', r'<a \1|\2>', text)
    # Replace remnants of \const; \type; \token, \norm; or \key; as pseudo xHTML <const/> tags
    text = re.sub(r'\\(const|type|token|norm|key);', r'<\1/>', text)
    # Put \c;Code\n; sequences into pseudo-HTML <code> tags
    text = re.sub(r'\\c;([^\\;]*?)\\n;', r'<code>\1</code>', text)
    # Replace remnants of \s; \c; \b; or \n; as pseudo xHTML <s/> tags
    text = re.sub(r'\\([scbn]);', r'<\1/>', text)
    return text

def convert_tag_syntax_to_escape_syntax(text):
    # Invert the replace remnants of \s; \c; \b; or \n; as pseudo xHTML <s/> tags
    text = re.sub(r'<([scbn])/>', r'\\\1;', text)
    # Invert the put of \c;Code\n; sequences into pseudo-HTML <code> tags
    text = re.sub(r'<code>([^\\;]*?)</code>', r'\\c;\1\\n;', text)
    # Invert the replace remnants of \const; \type; \token or \norm; as pseudo xHTML <const/> tags
    text = re.sub(r'<(const|type|token|norm|key)/>', r'\\\1;', text)
    # Invert the cleanup of pseudo-html targets separated by \\ to have a single character |
    text = re.sub(r'<a (.*?)\|(.*?)>', r'<a \1\\\2>', text)
    # Invert the transform of CBot links \l;text\u target; into pseudo-HTML <a target>text</a>
    text = re.sub(r'<a (.*?)>(.*?)</a>', r'\\l;\2\\u \1;', text)
    # Invert the put \const;Code\norm; sequences into pseudo-HTML <format const> tags
    text = re.sub(r'<format (const|type|token|key)>([^\\;]*?)</format>', r'\\\1;\2\\norm;', text)
    # Invert the replace of \button $id; as pseudo xHTML <button $id/> tags
    text = re.sub(r'<(button|key) (.*?)/>', r'\\\1 \2;', text)
    return text

"""
    Create jobs for help translation

    Assumes that input_dir has structure like so:
        ${input_dir}/E/help_file1.txt
        ...
        ${input_dir}/E/help_fileN.txt

    The output files will be saved in:
        ${output_dir}/${language_char1}/${install_subdir}/help_file1.txt
        ...
        ${output_dir}/${language_charM}/${install_subdir}/help_fileN.txt
"""
def create_help_translation_jobs(input_dir, output_dir, install_subdir, template_file, language_files):
    translation_jobs = []

    e_dir = os.path.join(input_dir, 'E')
    input_files = sorted(os.listdir(e_dir))

    if not install_subdir:
        install_subdir = ''

    language_files_list = []
    if len(language_files) > 0:
        language_files_list = language_files
    else:
        # We need at least one dummy language file to create any jobs
        language_files_list = [None]

    for language_file in language_files_list:
        output_translation_dir = None
        if language_file:
            output_translation_dir = nice_path_join(output_dir, language_file.language_char(), install_subdir)
            nice_mkdir(output_translation_dir)

        for input_file in input_files:
            translation_jobs.append(HelpTranslationJob(
                input_file    = os.path.join(e_dir, input_file),
                output_file   = nice_path_join(output_translation_dir, input_file),
                template_file = template_file,
                language_file = language_file))

    return translation_jobs



import errno
import io
import os
import polib

"""
   Conversion functions to avoid mixed \ and / path separators under Windows
"""
def convert_input_path(slash_path):
    if not slash_path:
        return None
    return slash_path.replace('/', os.sep)

def convert_output_path(system_path):
    if not system_path:
        return None
    return system_path.replace(os.sep, '/')

"""
    Works like shell's "mkdir -p" and also behaves nicely if given None argument
"""
def nice_mkdir(path):
    if path is None:
        return

    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise

"""
    Works as os.path.join, but behaves nicely if given None argument
"""
def nice_path_join(*paths):
    for path in paths:
        if path is None:
            return None

    return os.path.join(*paths)

"""
    Wrapper class over POFile, acting as translation template file

    It actually hold two POFile instances:
        previous_catalog is the content of PO file read from disk
        current_catalog is created empty and filled with entries from input files

    Once all processing is done, the content of previous_catalog is merged with current_catalog
    and the result is saved to disk.
"""
class TemplateFile:
    def __init__(self, file_name):
        self.file_name = file_name
        self.dir_name = os.path.dirname(file_name)
        self.language = 'en'
        self.current_catalog = polib.POFile(wrapwidth = 0)
        if os.path.exists(file_name):
            self.previous_catalog = polib.pofile(file_name, wrapwidth = 0)
        else:
            self.previous_catalog = polib.POFile(wrapwidth = 0)

    """
        Wrapper over inserting template file entry
        If entry does not exist, it is created;
         otherwise it is modified to indicate multiple occurrences
    """
    def insert_entry(self, text, occurrence, type_comment):
        entry = self.current_catalog.find(text)
        relative_file_name = convert_output_path(os.path.relpath(occurrence.file_name, self.dir_name))
        occurrence = (relative_file_name, occurrence.line_number)
        if entry:
            entry.comment = self._merge_comment(entry.comment, type_comment)
            if occurrence not in entry.occurrences:
                entry.occurrences.append(occurrence)
        else:
            comment = 'type: ' + type_comment
            new_entry = polib.POEntry(msgid = text,
                                      comment = comment,
                                      occurrences = [occurrence],
                                      flags = ['no-wrap'])

            self.current_catalog.append(new_entry)

    def _merge_comment(self, previous_comment, type_comment):
        new_comment = previous_comment

        previous_types = previous_comment.replace('type: ', '')
        previous_types_list = previous_types.split(', ')

        if type_comment not in previous_types_list:
            new_comment += ', ' + type_comment

        return new_comment

    """
        Merges previous_catalog with current_catalog and saved the result to disk
    """
    def merge_and_save(self):
        self.previous_catalog.merge(self.current_catalog)
        for x in self.previous_catalog.obsolete_entries():
            self.previous_catalog.remove(x)
        self.previous_catalog.save(self.file_name, newline='\n')

"""
    Wrapper class over POFile, acting as language translation file
"""
class LanguageFile:
    def __init__(self, file_name):
        self.file_name = file_name
         # get language from file name e.g. "/foo/de.po" -> "de"
        (self.language, _) = os.path.splitext(os.path.basename(file_name))
        if os.path.exists(file_name):
            self.catalog = polib.pofile(file_name, wrapwidth = 0)
        else:
            self.catalog = polib.POFile(wrapwidth = 0)

    """
        Return single language character e.g. "de" -> "D"
    """
    def language_char(self):
        if self.language == 'pt':
            return 'B';
        else:
            return self.language[0].upper()

    """
        Try to translate given text; if not found among translations,
        return the original
    """
    def translate(self, text):
        entry = self.catalog.find(text)
        if entry and entry.msgstr != '':
            return entry.msgstr
        return text

    """
        Merges entries with current_catalog from template file and saves the result to disk
    """
    def merge_and_save(self, template_file):
        self.catalog.merge(template_file.current_catalog)
        for x in self.catalog.obsolete_entries():
            self.catalog.remove(x)
        self.catalog.save(self.file_name, newline='\n')

"""
    Locates the translation files in po_dir
"""
def find_translation_file_names(po_dir):
    pot_file_name = os.path.join(po_dir, 'translations.pot') # default
    po_file_names = []
    for file_name in os.listdir(po_dir):
        if file_name.endswith('.pot'):
            pot_file_name = os.path.join(po_dir, file_name)
        elif file_name.endswith('.po'):
            po_file_names.append(os.path.join(po_dir, file_name))

    return (pot_file_name, po_file_names)

"""
    Creates template and language files by reading po_dir
"""
def create_template_and_language_files(po_dir):
    (pot_file_name, po_file_names) = find_translation_file_names(po_dir)

    template_file = TemplateFile(pot_file_name)
    language_files = []
    for po_file_name in po_file_names:
        language_files.append(LanguageFile(po_file_name))

    return (template_file, language_files)

"""
    Structure representing occurrence of text
"""
class Occurrence:
    def __init__(self, file_name, line_number):
        self.file_name = file_name
        self.line_number = line_number

"""
    Structure representing line read from input file
"""
class InputLine:
    def __init__(self, text, occurrence):
        self.text = text
        self.occurrence = occurrence


"""
    Base class for single translation process,
    translating one input file into one output file

    It provides wrapper code for reading consecutive lines of text and saving the result
"""
class TranslationJob:
    def __init__(self, **kwargs):
        self._input_line_counter = 0
        self._input_file_name = kwargs['input_file']
        self._input_file = None

        self._output_file_name = kwargs['output_file']
        self._output_file = None

    """
        Launch translation process
        Actual processing is done in process_file() function which must be implemented by subclasses
    """
    def run(self):
        try:
            self._open_files()
            self.process_file()
        finally:
            self._close_files()

    def _open_files(self):
        self._input_file = io.open(self._input_file_name, 'r', encoding='utf-8', newline='\n')
        if self._output_file_name:
            self._output_file = io.open(self._output_file_name, 'w', encoding='utf-8', newline='\n')

    def _close_files(self):
        self._input_file.close()
        if self._output_file:
            self._output_file.close()

    """
        Return next line, occurrene pair from input file or None if at end of input
    """
    def read_input_line(self):
        line = self._input_file.readline()
        if line == '':
            return None

        self._input_line_counter += 1
        return InputLine(line.rstrip('\n'), Occurrence(self._input_file_name, self._input_line_counter))

    """
        Write line to output file, if present
    """
    def write_output_line(self, line):
        if self._output_file:
            self._output_file.write(line + '\n')

    def get_input_file_name(self):
        return self._input_file_name

    def get_output_file_name(self):
        return self._output_file_name



import os

from translate_level import LevelTranslationJob
from common import nice_mkdir, nice_path_join

"""
    Create jobs for chaptertile files translation

    Assumes that input_dir has structure like so:
        ${input_dir}/dir1/chaptertitle.txt
        ...
        ${input_dir}/dirN/chaptertitle.txt

    The output files will be saved in:
        ${input_dir}/dir1/chaptertitle.txt
        ...
        ${input_dir}/dirN/chaptertitle.txt

    The actual translation is done using the same jobs as level files
"""
def create_chaptertitles_translation_jobs(input_dir, output_dir, template_file, language_files):
    translation_jobs = []

    for subdirectory in sorted(os.listdir(input_dir)):
        input_subdirectory = os.path.join(input_dir, subdirectory)
        if not os.path.isdir(input_subdirectory):
            continue

        input_file = os.path.join(input_subdirectory, 'chaptertitle.txt')
        if not os.path.isfile(input_file):
            continue

        output_subdirectory = nice_path_join(output_dir, subdirectory)
        nice_mkdir(output_subdirectory)

        translation_jobs.append(LevelTranslationJob(
            input_file     = input_file,
            output_file    = nice_path_join(output_subdirectory, 'chaptertitle.txt'),
            template_file  = template_file,
            language_files = language_files))

    return translation_jobs



import os
import re

from translate_help import HelpTranslationJob
from common import TranslationJob, nice_mkdir, nice_path_join

class LevelTranslationJob(TranslationJob):
    def __init__(self, **kwargs):
        TranslationJob.__init__(self, **kwargs)
        self.template_file = kwargs['template_file']
        self.language_files = kwargs['language_files']
        self.line_buffer = None

    def process_file(self):
        while True:
            line = self.read_input_line()
            if not line:
                break

            # English version is always written
            self.write_output_line(line.text)

            match = re.match(r'^(Title|Resume|ScriptName)\.E(.*)', line.text)
            if match:
                for language_file in self.language_files:
                    self.add_translated_line(match, line.occurrence, language_file)

    def add_translated_line(self, command_match, occurrence, language_file):
        command = command_match.group(1)
        arguments = command_match.group(2)

        translated_arguments = arguments
        for attribute_match in re.finditer('(text|resume)="([^"]*)"', arguments):
            whole_attribute_match = attribute_match.group(0)
            attribute = attribute_match.group(1)
            text = attribute_match.group(2)

            self.template_file.insert_entry(text, occurrence, command + '-' + attribute)

            translated_arguments = translated_arguments.replace(
                whole_attribute_match,
                u'{0}="{1}"'.format(attribute, language_file.translate(text)))

        self.write_output_line(u'{0}.{1}{2}'.format(
                command,
                language_file.language_char(),
                translated_arguments))


"""
    Create jobs for chapter translation

    Assumes that input_dir has structure like so:
        ${input_dir}/scene.txt
        ${input_dir}/help/help_file1.txt
        ...
        ${input_dir}/help/help_fileN.txt

    The output files will be saved in:
        ${output_dir}/scene.txt
        ${output_dir}/help/help_file1.${language_char1}.txt
        ...
        ${output_dir}/help/help_fileN.${language_charM}.txt
"""
def create_level_translation_jobs(input_dir, output_dir, template_file, language_files):
    translation_jobs = []

    input_file = os.path.join(input_dir, 'scene.txt')
    translation_jobs.append(LevelTranslationJob(
        input_file     = input_file,
        output_file    = nice_path_join(output_dir, 'scene.txt'),
        template_file  = template_file,
        language_files = language_files))

    input_help_dir = os.path.join(input_dir, 'help')
    if os.path.isdir(input_help_dir):
        output_help_dir = nice_path_join(output_dir, 'help')
        nice_mkdir(output_help_dir)

        language_files_list = []
        if len(language_files) > 0:
            language_files_list = language_files
        else:
            # We need at least one dummy language file to create any jobs
            language_files_list = [None]

        for language_file in language_files_list:
            for help_file in sorted(os.listdir(input_help_dir)):
                if language_file:
                    translated_help_file = help_file.replace('.E.txt', '.{0}.txt'.format(language_file.language_char()))

                translation_jobs.append(HelpTranslationJob(
                    input_file    = os.path.join(input_help_dir, help_file),
                    output_file   = nice_path_join(output_help_dir, translated_help_file),
                    template_file = template_file,
                    language_file = language_file))

    return translation_jobs
